/*****************************************************************************/
/*								             */
/* Author: Michael C. Nechyba					             */
/* Last modified: 9/20/98                   			             */
/*								             */
/* File: "bp_main.h" 						             */
/*								             */
/* This code is modified from the freely available NevProp v1.16 neural      */
/* network simulator (authored by Phil Goodman, David Rosen and Allen        */
/* Plummer).                                                                 */
/*								             */
/*****************************************************************************/

/******************************* Global variables ****************************/

char  *NetworkFile;        /* Nechyba change */
char  fname[256];          /* global .net prefix */
int   Epoch;               /* Current epoch number */
float WtRange;             /* Random-init weights in range [-WR,+WR] */
float SigmoidPrimeOffset;  /* Add to sigmoid-prime to kill flat spots */
int   HyperErr;            /* 1=> use atanh error function */
int   SplitEpsilon;        /* 1=> divide epsilon by fan-in before use */
float Epsilon;             /* For grad descent if last step was (almost) 0 */
float Momentum;            /* Normal old momentum term */
float ModeSwitchThreshold; /* Inside thresh, do grad descent; outside, jump. */
float MaxFactor;           /* Don't jump more than this times last step */
float Decay;               /* Weight decay */
int   KeepScore;           /* 1=> accumulate error score for each epoch */
float TotalError;          /* Total output error for one epoch */
float ScoreThreshold;      /* This close to desired value=> bit is correct */
int   TotalErrorBits;      /* Total # bits in epoch that were wrong */
int   DidGradient;         /* Total # patterns that did gradient descent */
int   Nunits;              /* Total number of units in net */
int   Ninputs;             /* Number of input units */
int   FirstHidden;         /* Index of 1st hidden unit */
int   Nhidden;             /* Number of hidden units */
int   FirstOutput;         /* Index of 1st output unit */
int   Noutputs;            /* Number of output units */
int   NTrainingPatterns; 
int   NTestingPatterns;  
FILE  *resfile;         /* Send results to log file */
FILE  *errfile;         /* Send results to err file */
unsigned long seed;
int   ReportFreq;       /* #epochs between reports - added 2/20/93 phg */
int   TrainErrorBits;   /* storage for TRAIN reports */
float TrainSSE;         /* storage for TRAIN reports */
float TrainRMS;
int   TestErrorBits;    /* storage for TESTING reports - added 2/21/93 phg */
float TestSSE;          /* storage for TESTING reports - added 2/21/93 phg */
float TestRMS;
int   UseQuickProp;     /* 0=> forces gradient descent always */
int   WeightCount;      /* counts connections, reporting wt update by method */
int   ReportRandSeed;   /* regurgitate only if weights randomized */
int   EpochWiseUpdate;  /* switches betw per epoch and per pattern update */
int   MinEpochs;        /* min epochs before stopped training by test error */
float BeyondBestEpoch;  /* Stop training when Epoch no. exceeds best Epoch
                              by this multiple */
int   EpochSaved;       /* Epoch at which a given .wts file was frozen */ 
int   ForceTraining;    /* 1 => override stop-training at end of TRAIN() */
int   PredictOnly;      /* 1 => do Epoch 0 only to make predictions only */

int   BestEpoch;
float BestTestRMS;
int   SaveBestWts;

int   **Connections;    /* C[i][j] lists jth unit projecting to unit i */
float **Weights;        /* W[i][j] holds weight of C[i][j] */
float **DeltaWeights;   /* Change between previous weight and current one */
float **Slopes;         /* Accumulated slope value for each position */
float **PrevSlopes;     /* Similarly, for the last position visited */
float **TrainingInputs; /* Double pointer for array accessing */
float **TrainingOutputs;
float **TestInputs;
float **TestOutputs;
float **BestWeights;    /* Holds weights of best reported test C[i][j] */
float *Outputs;         /* Final output value for each unit */
float *ErrorSums;       /* Total error activation for each unit */
int   *Nconnections;    /* # of INCOMING connections per unit */

/******************************* Nechyba globals *****************************/

char TrainingDataFile[256];
char ValidationDataFile[256];
char ExistingWeightFile[256];
char realerrfname[256], realresfname[256];

int  NParameters;
int  PredictOnly;
int  RandomSeed;
int  RetainBestWeights;
int  MaxEpochs;
int  ProtectFiles;
int  SuppressConsole;
int  RecordFreq;
int  NTrials;
int  Trial;

/* 
 * Unit Types:
 *
 * 0 --> Linear
 * 1 --> Sigmoid (-0.5 to 0.5)
 * 2 --> Asymmetric sigmoid ( 0.0 to 1.0)
 *
 */

int  HiddenUnitType;
int  OutputUnitType;

PARAMETERS ParameterTable[] = {
  {"SuppressConsole",       INT, (POINTER) &SuppressConsole},
  {"ProtectFiles",          INT, (POINTER) &ProtectFiles},
  {"UseQuickProp",          INT, (POINTER) &UseQuickProp},
  {"EpochWiseUpdate",       INT, (POINTER) &EpochWiseUpdate},
  {"MaxEpochs",             INT, (POINTER) &MaxEpochs},
  {"ReportFreq",            INT, (POINTER) &ReportFreq},
  {"RecordFreq",            INT, (POINTER) &RecordFreq},
  {"MinEpochs",             INT, (POINTER) &MinEpochs},
  {"BeyondBestEpoch",       FLT, (POINTER) &BeyondBestEpoch},
  {"WeightRange",           FLT, (POINTER) &WtRange},
  {"HyperErr",              INT, (POINTER) &HyperErr},
  {"SigmoidPrimeOffset",    FLT, (POINTER) &SigmoidPrimeOffset},
  {"Epsilon",               FLT, (POINTER) &Epsilon},
  {"SplitEpsilon",          INT, (POINTER) &SplitEpsilon},
  {"Momentum",              FLT, (POINTER) &Momentum},
  {"WeightDecay",           FLT, (POINTER) &Decay},
  {"ScoreThreshold",        FLT, (POINTER) &ScoreThreshold},
  {"QPMaxFactor",           FLT, (POINTER) &MaxFactor},
  {"QPModeSwitchThreshold", FLT, (POINTER) &ModeSwitchThreshold},
  {"TrainingFile",          STR, (POINTER) &TrainingDataFile},
  {"ValidationFile",        STR, (POINTER) &ValidationDataFile},
  {"HiddenUnitType",        INT, (POINTER) &HiddenUnitType},
  {"OutputUnitType",        INT, (POINTER) &OutputUnitType},
  {"ExistingWeightFile",    STR, (POINTER) &ExistingWeightFile},
  {"PredictOnly",           INT, (POINTER) &PredictOnly},
  {"RandomSeed",            INT, (POINTER) &RandomSeed},
  {"RetainBestWeights",     INT, (POINTER) &RetainBestWeights}

  /*
    Two other parameters are handled specially by the parameter
     file routines in bp_io.c:
     
     Configuration <ninputs> <nhidden> <noutputs>
     ConnectCall <c1> <c2> <c3> <c4>

     Note that the ConnectCall argument may be repeated. However,
     it MUST come after the ConfigurationCall.
  */
};
