/*****************************************************************************/
/*								             */
/* Author: Michael C. Nechyba					             */
/* Last modified: 11/5/2002                   			             */
/*								             */
/* File: "nn_util.c" 						             */
/*								             */
/* This code is modified from the freely available NevProp v1.16 neural      */
/* network simulator (authored by Phil Goodman, David Rosen and Allen        */
/* Plummer).                                                                 */
/*								             */
/*****************************************************************************/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <assert.h>
#include "nn_util.h"
#include "nn_main.h"

#include "assert.h"
#ifdef DEBUG
#define dbassert(expression) assert(expression)
#else
#define dbassert(expression)
#endif

/**************************** Subroutines ************************************/

int main_loop(void)
{
  int epox, epoxtrue = 0;
  char *starttime, *completedtime;
  time_t now;
  struct tm *date;

  now = time( NULL );
  date = localtime( &now );
  starttime = asctime( date );
  
  epox = 1;  /* arbitrary positive */
  
  INITIALIZE_GLOBALS();
  GET_NETWORK_CONFIGURATION();

  if (NTrainingPatterns < 1 && NTestingPatterns < 1) {
    fprintf(stderr, NOPATTERNS);  
    exit(1);
  }

  /* Get result file name */

  protected_file(realresfname, fname, RES_EXTENSION, ProtectFiles);
  resfile = fopen(realresfname, "w");

  protected_file(realerrfname, fname, ERR_EXTENSION, ProtectFiles);
  errfile = fopen(realerrfname, "w");

  fprintf(resfile, 
	  "\n\n######### Welcome to NevProp ##########\n");
  fprintf(resfile, "NevProp started on %s", starttime);
 
  if(file_exist(ExistingWeightFile)) {
    ReportRandSeed = 0;
    GET_WEIGHTS();

    if (NTestingPatterns > 0 && PredictOnly > 0) {
      fprintf(stderr, 
	      "Using existing weight file to predict validation set.\n");

      MAKE_PREDICTIONS(); /* which will exit the program */
    }
  } else {
    PredictOnly = 0;
    
    if(RandomSeed < 0) {
      seed = time((void *)0); 
    } else {
      if(NTrials > 1) {
	seed = RandomSeed + Trial - 1;
      } else {
	seed = RandomSeed;
      }
    }

    srand(seed);
    INIT_WEIGHTS();
  }

  if (NTestingPatterns > 0 && NTrainingPatterns > 0) {
    if(RetainBestWeights > 0) {
      SaveBestWts = 1;

      BestEpoch = 1;
      BestTestRMS = 1000000000.;
    } else {
      SaveBestWts = 0;
    }
  }

  /* Train the sucker. */

  while (epox != 0 && NTrainingPatterns > 0) {
    if (Epoch==0 || Epoch==EpochSaved) {
      epox = MaxEpochs;

      if (epox == 0 && NTestingPatterns > 0) {
	MAKE_PREDICTIONS(); /* which will then exit the program */
      }

      epoxtrue=epox; /* used for error check and default epox if continuing */

      epox=epox + 1;

    } else {
      PRINT_BEST(); /* display the little histogram at this point */
      
      epox=0;
    }

    if (epox !=0) {
      if (ReportFreq > epoxtrue || ReportFreq < 0) {
	ReportFreq = epoxtrue;
      }
      
      if(ReportFreq == 0) {
	ReportFreq = 1;
      }
      
      if (Epoch==0) {
#define printargs "\n*****************************************\n"
	printboth;
#undef printargs

#define printargs "Trial # %d started at %s\n", Trial, starttime
	printboth;
#undef printargs 

	if(SuppressConsole > 0) {
	  fprintf(stderr, "Trial # %d started at %s\n", Trial, starttime);
	}
      }
      
      if(SuppressConsole == 0) {
	PRINT_GLOBALS(stdout, fname);  /* Regurgitate global variables */
      }
      
      if (resfile) {
	PRINT_GLOBALS(resfile, fname);
      }

      TRAIN(epox);
    }
  }
    
  TEST_SAVE();
  DUMP_WEIGHTS();

  if (resfile) {
    now = time( NULL );
    date = localtime( &now );
    completedtime = asctime( date );
    
#define printargs "\nTraining log saved to \"%s\"\n\n", realresfname
  printboth;
#undef printargs

#define printargs "RMS error over time saved to \"%s\"\n", realerrfname
  printboth;
#undef printargs

#define printargs "\n*****************************************\n"
    printboth;
#undef printargs

#define printargs "Trial # %d completed at %s\n", Trial, completedtime
    printboth;
#undef printargs 

    if(SuppressConsole > 0) {
      fprintf(stderr, "Trial # %d completed at %s\n", Trial, completedtime);
    }
  }
  
  fclose(resfile);
  fclose(errfile);
  
  FREE_MEMORY();
  return(1);
}

/*****************************************************************************/

void INITIALIZE_GLOBALS(void) 
{
  HiddenUnitType = SIGMOID;
  OutputUnitType = LINEAR;
  SuppressConsole = 0;
  UseQuickProp = 0;
  EpochWiseUpdate = 0;
  MinEpochs = 200;
  BeyondBestEpoch = 2.;
  WtRange = 1.; /* Corresponds to WeightRange */
  HyperErr = 0;
  SigmoidPrimeOffset = 0.;
  Epsilon = 0.1;
  SplitEpsilon = 0;
  Momentum = 0.;
  Decay = 0.; /* Corresponds to WeightDecay */
  ScoreThreshold = 0.01;
  MaxFactor = 1.75; /* Corresponds to QPMaxFactor */
  ModeSwitchThreshold = 0.0; /* Corresponds to QPModeSwitchThreshold */

  TrainingDataFile[0] = 0;
  ValidationDataFile[0] = 0;
  ExistingWeightFile[0] = 0;
  PredictOnly = 0;
  RandomSeed = -1;
  RetainBestWeights = 1;
  MaxEpochs = 1000;
  ReportFreq = 1;
  RecordFreq = 1;
  ProtectFiles = 0;

  Epoch=0;
  KeepScore=0.0;  /* intermittenlty gets reset to 1 for reporting  */
  TotalError=0.0;
  ScoreThreshold=0.35;
  TotalErrorBits=0;

  /* set the seed to the current time for default setting */

  SaveBestWts = 1;
  DidGradient = 0;
  WeightCount = 0;
  ReportRandSeed = 1;

  EpochSaved = 0;
  ForceTraining = 0;


  resfile = NULL;
  errfile = NULL;
}

/*****************************************************************************/

void GET_NETWORK_CONFIGURATION(void) 
{
  FILE  *infile;
  char realfname[256], tmp[256], *first;
  int  i, j;
  float tempfloat, *floatPtr;
  
  /* Nechyba change */
  
  sprintf(realfname, "%s.net", NetworkFile);
  
  if (!file_exist(realfname)) {
    
    if(!file_exist(NetworkFile)) {
      fprintf(stderr,"That network file wasn't found; exiting.\n");
      exit(1);
    } else {
      sprintf(realfname, "%s", NetworkFile);

      sprintf(tmp, "%s", NetworkFile);
      first = strtok(tmp, ".");
      sprintf(fname, "%s", first);
    }
  } else {
    sprintf(fname, "%s", NetworkFile);
  }

  NParameters = sizeof(ParameterTable)/sizeof(PARAMETERS);
  get_parameters(realfname, ParameterTable, NParameters);

  if(HiddenUnitType == 0) {
#define printargs "WARNING: Hidden unit types are not allowed to be linear.\n"
    printboth;
#undef printargs

#define printargs "Defaulting to simgoidal (-.5 to .5) activation functions.\n"
    printboth;
#undef printargs 
    HiddenUnitType = SIGMOID;
  }

  if(NTrials > 1) {
    ProtectFiles = 1;
  }

  if (UseQuickProp == 1 && EpochWiseUpdate == 0) {
    EpochWiseUpdate = 1;
    fprintf(stderr, "WARNING: EpochWiseUpdate was set to 0 in %s.net.\n",
	    fname);
    fprintf(stderr, "Quickprop, however, requires EpochWiseUpdate to\n");
    fprintf(stderr, "be 1. Therefore EpochWiseUpdate is being reset to 1.\n");
  }

  /* Read in number of training patterns */

  NTrainingPatterns = count_lines(TrainingDataFile);
  NTestingPatterns = count_lines(ValidationDataFile);

  if(NTrainingPatterns <= 0) {
    fprintf(stderr, "ERROR: Could not find training data file; exiting.\n");
    exit(1);
  }

  if(NTestingPatterns <= 0) {
    fprintf(stderr, 
	    "WARNING: Could not find validation data file.\n");
    fprintf(stderr, 
	    "WARNING: Will default validation data to training data.\n");
    strcpy(&ValidationDataFile[0], &TrainingDataFile[0]);
  }

  /* Now get memory for training and testing patterns */
  TrainingInputs=(float **)malloc(((NTrainingPatterns +1) * 
				   sizeof(floatPtr)));
  assert(TrainingInputs);
  for(i=0;i<NTrainingPatterns;i++) {
    TrainingInputs[i]=(float *)malloc(((Ninputs+1) * sizeof(float)));
    assert(TrainingInputs[i]);
  }

  TrainingOutputs=(float **)malloc(((NTrainingPatterns+1) * 
				    sizeof(floatPtr)));
  assert(TrainingOutputs);
  for(i=0;i<NTrainingPatterns;i++) {
    TrainingOutputs[i]=(float *)malloc(((Noutputs+1) * sizeof(float)));
    assert(TrainingOutputs[i]);
  }

  TestInputs=(float **)malloc(((NTestingPatterns+1) * sizeof(floatPtr)));
  assert(TestInputs);
  for(i=0;i<NTestingPatterns;i++) {
    TestInputs[i]=(float *)malloc(((Ninputs+1) * sizeof(float)));
    assert(TestInputs[i]);
  }

  TestOutputs=(float **)malloc(((NTestingPatterns+1) * sizeof(floatPtr)));
  assert(TestOutputs);
  for(i=0;i<NTestingPatterns;i++) {
    TestOutputs[i]=(float *)malloc(((Noutputs+1) * sizeof(float)));
    assert(TestOutputs[i]);
  }

  infile = fopen(TrainingDataFile, "r");

  /* Read patterns into their respective memory blocks */
  for (i=0; i<NTrainingPatterns; i++) {
    for (j=0; j<Ninputs; j++) {
      checkEOF(fscanf(infile, "%f", &tempfloat));
      TrainingInputs[i][j]=tempfloat;
    }
    for (j=0; j<Noutputs; j++) {
      checkEOF(fscanf(infile, "%f", &tempfloat));
      TrainingOutputs[i][j]=tempfloat;
    }
  }

  fclose(infile);
  infile = fopen(ValidationDataFile, "r");

  /* Read in the test patterns */
  for (i=0; i<NTestingPatterns; i++) {
    for (j=0; j<Ninputs; j++)  {
      checkEOF(fscanf(infile, "%f", &tempfloat));
      TestInputs[i][j]=tempfloat;
    }
    for (j=0; j<Noutputs; j++) {
      checkEOF(fscanf(infile, "%f", &tempfloat));
      TestOutputs[i][j]=tempfloat;
    }
  }

  fclose(infile);
 
  fflush(stdout);
}

/*****************************************************************************/

void BUILD_DATA_STRUCTURES(int ninputs, int nhidden, int noutputs)
{
  int i, *intPtr;
  float *floatPtr;
  
  Nunits      = 1 + ninputs + nhidden + noutputs;
  Ninputs     = ninputs;
  FirstHidden = 1 + ninputs;
  Nhidden     = nhidden;
  FirstOutput = 1 + ninputs + nhidden;
  Noutputs    = noutputs;

  Connections = (int **)malloc((Nunits * sizeof(intPtr)));
  assert(Connections);

  for(i = 0;i < Nunits;i++) {
    Connections[i] = (int *)malloc((Nunits * sizeof(int)));
    assert(Connections[i]);
  }

  Weights = (float **)malloc((Nunits * sizeof(floatPtr)));
  assert(Weights);

  for(i = 0;i < Nunits;i++) {
    Weights[i] = (float *)malloc((Nunits * sizeof(float)));
    assert(Weights[i]);
  }

  BestWeights = (float **)malloc((Nunits * sizeof(floatPtr)));
  assert(BestWeights);

  for(i = 0;i < Nunits;i++) {
    BestWeights[i] = (float *)malloc((Nunits * sizeof(float)));
    assert(BestWeights[i]);
  }

  DeltaWeights = (float **)malloc((Nunits * sizeof(floatPtr)));
  assert(DeltaWeights);

  for(i = 0;i < Nunits;i++) {
    DeltaWeights[i] = (float *)malloc((Nunits * sizeof(float)));
    assert(DeltaWeights[i]);
  }

  Slopes = (float **)malloc((Nunits * sizeof(floatPtr)));
  assert(Slopes);

  for(i = 0;i < Nunits;i++) {
    Slopes[i] = (float *)malloc((Nunits * sizeof(float)));
    assert(Slopes[i]);
  }

  PrevSlopes = (float **)malloc((Nunits * sizeof(floatPtr)));
  assert(PrevSlopes);

  for(i = 0;i < Nunits;i++) {
    PrevSlopes[i] = (float *)malloc((Nunits * sizeof(float)));
    assert(PrevSlopes[i]);
  }

  Outputs = (float *)malloc((Nunits * sizeof(float)));
  assert(Outputs);
  
  ErrorSums = (float *)malloc((Nunits * sizeof(float)));
  assert(ErrorSums);
  
  Nconnections = (int *)malloc((Nunits * sizeof(int)));
  assert(Nconnections);

  for (i = 0; i < Nunits; i++)
    Outputs[i] = 0.0;

  for (i = 0; i < Nunits; i++)
    Nconnections[i] = 0;

  Outputs[0] = 1.0;        /* The bias unit */
}

/*****************************************************************************/

void CONNECT_LAYERS(int start1, int end1, int start2, int end2)
{
  int i, j, k;

  for (i = start2; i <= end2; i++) {
    if(Nconnections[i] == 0) {
      Nconnections[i]  +=1;
      Connections[i][0] = 0;
      DeltaWeights[i][0] = 0.0;
      Slopes[i][0] = 0.0;
      PrevSlopes[i][0] = 0.0;
      k=1;
    } else {
      k=Nconnections[i];
    }

    for (j = start1; j <= end1; j++) {
      Nconnections[i]  += 1;
      Connections[i][k] = j;
      DeltaWeights[i][k] = 0.0;
      Slopes[i][k] = 0.0;
      PrevSlopes[i][k] = 0.0;
      k++;
    }
  }
}

/*****************************************************************************/

void GET_WEIGHTS(void) 
{
  FILE  *infile;
  int  i, j, k, l;
  float inweight;
  char realfname[280];

  sprintf(realfname, "%s.wts", ExistingWeightFile);
 
  if(!file_exist(realfname)) {
    sprintf(realfname, "%s", ExistingWeightFile);

    if(!file_exist(realfname)) {
      fprintf(stderr, WTSEXIST);
      exit(1);
    }
  }

  infile = fopen(realfname, "r");

  /* Read in Epoch at which the wts were frozen */
  checkEOF(fscanf(infile, "%*s %d", &EpochSaved));

  for (i=0; i<Nunits; i++)
    for (j=0; j<Nunits; j++)  
      Weights[i][j]=0.0;    /* Zeroed weights */

  /* Read in Weights based on .net file construction */

  for (j=FirstHidden; j<Nunits; j++) {
    for (i=0; i<Nconnections[j]; i++) {
      checkEOF(fscanf(infile, "%d %d %f", &k, &l, &inweight));
      /* Check array elements for proper placement of weight values */
      if ((j==k) && (l==Connections[j][i]))  Weights[j][i]=inweight;
      assert(l==Connections[j][i]);
    }
  }
  Epoch=EpochSaved;

  fclose (infile);

#define printargs "Weights retrieved from \"%s\"\n", realfname
  printboth;
#undef printargs
}

/*****************************************************************************/

void MAKE_PREDICTIONS(void) 
{
  TRAIN(1);
  TEST_SAVE(); 

  FREE_MEMORY();

  exit(0);
}

/*****************************************************************************/

void INIT_WEIGHTS(void) 
{
  int i, j;
  float x;

  for (i = 0; i < Nunits; i++) {
    for (j = 0; j < Nconnections[i]; j++) {
      x = RANDOM_WEIGHT(WtRange);

      if(x < -WtRange || x > WtRange) {
	fprintf(stderr, 
		"ERROR: RANDOM_WEIGHT() value out of range. Exiting.\n");
	exit(1);
      }

      Weights[i][j] = x;
      DeltaWeights[i][j] = 0.0;
      Slopes[i][j] = 0.0;
      PrevSlopes[i][j] = 0.0;
    }
  }
}

/*****************************************************************************/

float RANDOM_WEIGHT(float range)
{
  return((float) ((2.*range)*rand()/RAND_MAX - range));
}

/*****************************************************************************/

void PRINT_GLOBALS(FILE *stream, char *filename)
{
  fprintf(stream, "\nParameters read from file \"%s.net\"...\n\n",filename);
  fprintf(stream, "TRAINING set patterns: %d\n", NTrainingPatterns);
  fprintf(stream, "TESTING set patterns:  %d\n\n", NTestingPatterns);

  fprintf(stream, "InputUnits: %d -- HiddenUnits: %d -- OutputUnits: %d\n\n",
          Ninputs, Nhidden, Noutputs);

  if (ReportRandSeed == 1) {
    fprintf(stream, "SEED for initial random weights=%lu.\n\n", 
	    seed);
  }

  display_parameters(ParameterTable, NParameters, stream);
}

/*****************************************************************************/

void TRAIN(int times)
{
  int i, report, record;

  report = ReportFreq;
  record = RecordFreq;

  for (i = 0; i < times; i++) { 

    if (Epoch % record == 0) { /* Time to log to .err file */
      TRAIN_SUMMARY();
      TEST_SUMMARY();

      if (i > 0 || (Epoch == EpochSaved && Epoch > 0))
	CHECK_IF_BEST();

      fprintf(errfile, "%d %.6f %.6f\n", Epoch, TrainRMS, TestRMS);
    }

    if (Epoch % report == 0) { /* Time to report status */
      TRAIN_SUMMARY();
      TEST_SUMMARY();

      if (i > 0 || (Epoch == EpochSaved && Epoch > 0))
	CHECK_IF_BEST();

      if (Epoch==0 || Epoch==EpochSaved || i > 0) {
#define printargs "*----------------------------------------------------*\n"
	printboth; 
#undef printargs
	if (Epoch==0 || Epoch==EpochSaved) {
#define printargs "Epoch %d: \n", Epoch
	  printboth; 
#undef printargs
	} else {
#define printargs "Epoch %d: QuickProp on %5.2f%%, GradDesc on %5.2f%%.\n",\
	  Epoch, 100.0*(1.0 - (float)DidGradient/WeightCount),\
	    100.0*(float)DidGradient/WeightCount
	    printboth; 
#undef printargs
	}
    
	/* display the results */

	if (NTrainingPatterns > 0)
	  PRINT_RESULTS("TRAINING", 
			NTrainingPatterns, TrainErrorBits, TrainSSE);

	if (NTestingPatterns > 0)
	  PRINT_RESULTS(" TESTING", 
			NTestingPatterns, TestErrorBits, TestSSE);

	fflush(stdout);
      }

      if (!STOPPED_TRAINING() && i < times-1) {
	KeepScore=1;

	TRAIN_ONE_EPOCH();

	KeepScore = 0;  /* Turn off unnec. calcs now */
      } else
	i = times;
    } else if(i < times - 1) {
      TRAIN_ONE_EPOCH();
    }
  }
}

/*****************************************************************************/

void TRAIN_SUMMARY(void)
{
  int j, k;

  KeepScore=1;
  TotalError=0.0;
  TotalErrorBits=0;
  
  for(j = 0; j < NTrainingPatterns; j++) {

    FORWARD_PASS(TrainingInputs[j]);

    for (k=FirstOutput; k<Nunits; k++) {

      (void) ERRFUN(TrainingOutputs[j][k-FirstOutput], Outputs[k]);
    }
  }

  TrainErrorBits=TotalErrorBits; /* Rename these for use later  -phg */
  TrainSSE = TotalError;
  TrainRMS = rms(TrainSSE, NTrainingPatterns, Noutputs);
  KeepScore=0;
}

/*****************************************************************************/

void TEST_SUMMARY(void)
{
  int j, k;

  KeepScore=1;
  TotalError=0.0;
  TotalErrorBits=0;

  for(j=0; j<NTestingPatterns; j++) {
    FORWARD_PASS(TestInputs[j]);

    for (k=FirstOutput; k<Nunits; k++) {
      (void) ERRFUN(TestOutputs[j][k-FirstOutput], Outputs[k]);
    }
  }

  TestErrorBits=TotalErrorBits; /* Rename these for use later  -phg */
  TestSSE = TotalError;
  TestRMS = rms(TestSSE, NTestingPatterns, Noutputs);
  KeepScore=0;
}

/*****************************************************************************/

void CHECK_IF_BEST(void)
{
  int m, n;

  if(BestTestRMS > TestRMS) {
    for (m = 0; m < Nunits; m++) {
      for (n = 0; n < Nunits; n++)  
	BestWeights[m][n]=Weights[m][n];
    }
    
    BestEpoch = Epoch;
    BestTestRMS = TestRMS;
  }
}

/*****************************************************************************/

/*
 *  TRAIN_ONE_EPOCH():
 *  Perform forward and back propagation once for each pattern in the
 *  training set, collecting deltas.  Then burn in the weights.
 *
 *  EpochWiseUpdate(per-epoch) vs per-pattern option offered.
 *  If using QuickProp, only EpochWiseUpdate available at present -
 *  check with Scott Fahlman for progress on this issue. -phg
 */

void TRAIN_ONE_EPOCH(void)
{
  int i;

  DidGradient=0;  /* Reinitialize prior to training another epoch  -phg*/
  TotalError=0.0;
  TotalErrorBits=0;

  if (EpochWiseUpdate) CLEAR_SLOPES();

  for (i=0; i<NTrainingPatterns; i++) {
    if (!EpochWiseUpdate) CLEAR_SLOPES();

    FORWARD_PASS (TrainingInputs[i]);
    BACKWARD_PASS (TrainingOutputs[i]);

    if (!EpochWiseUpdate) UPDATE_WEIGHTS();
  }

  if (EpochWiseUpdate) UPDATE_WEIGHTS();
  
  Epoch++;
}

/*****************************************************************************/

void FORWARD_PASS(float *input)
{
  int i, j, *connect, unit_type;
  float sum, *weight;

  /* Load in the input vector */

  for (i = 0; i < Ninputs; i++)  
    Outputs[i+1] = input[i];

  /* For each unit, collect incoming activation and pass through sigmoid. */

  for (j = FirstHidden; j < Nunits; j++) {
    connect = Connections[j];
    weight = Weights[j];
    sum = 0.0;
    i = Nconnections[j];

    while (i > 0) {
      i--;
      sum += (Outputs[ connect[i] ] * weight[i]);
    }

    
    if(j < FirstOutput)
      unit_type = HiddenUnitType;
    else
      unit_type = OutputUnitType;

    Outputs[j] = ACTIVATION(sum, unit_type);
  }
}

/*****************************************************************************/

/*
 *  ACTIVATION():
 *  Given the sum of weighted inputs, compute the unit's activation value.
 *  Defined unit types are SIGMOID and ASYMSIGMOID.
 */

float ACTIVATION(float sum, int unit_type)
{
  switch(unit_type) 
    {
    case LINEAR: 
      {
	return(sum);
      }
    case SIGMOID:
      {
	if (sum < -30.0)  
	  return(-0.5);  /* -30 used to prevent overflow */

	return(float)(1.0/(1.0 + (exp((double) - sum))) - 0.5);
      }
    default:
      {
	dbassert(unit_type==ASYMSIGMOID);

	/* asymmetrical sigmoid function in range 0.0 to 1.0. */

	if (sum < -30.0)  
	  return(0.0);  /* -30 used to prevent overflow */

	return(float)(1.0/(1.0 + exp((double) - sum)));
      }
    }
}

/*****************************************************************************/

/*
 *  ERRFUN():
 *  Compute the error for one output unit.
 *  For reporting periods, calculate SSE and error bits.
 *  HyperErr==0 => use squared error.
 *  HyperErr==1 => use tanh.
 */

float ERRFUN(float desired, float actual)
{
  float dif;

  dif = desired - actual;

  if (KeepScore) {
    TotalError +=dif*dif;

    if (((float) fabs((double) dif)) >=ScoreThreshold)
      TotalErrorBits++;
  }

  if (HyperErr==0)  
    return (dif); /* Not using atanh for error */
 else  
   return ((float) log ((double) (1.0+dif) / (1.0-dif)));
}

/*****************************************************************************/

/*
 *  STOPPED_TRAINING():
 */

int STOPPED_TRAINING(void)
{
  int stop_epoch;

  stop_epoch = (int) (BeyondBestEpoch * (float) BestEpoch);

  if(stop_epoch < Epoch) {
#define printargs "Stopping training early at epoch %d.\n", Epoch
	printboth;
#undef printargs

    return(1);
  } else {
    return(0);
  }
}

/*****************************************************************************/

/*
 *  CLEAR_SLOPES():
 *  Save the current slope values as PrevSlopes, and "clear" all current
 *  slopes (actually set to corresponding weight, decayed a bit).
 */

void CLEAR_SLOPES(void)
{
  int i, j;

  for (i = FirstHidden; i < Nunits; i++) {
    for (j=0; j<Nconnections[i]; j++) {
      PrevSlopes[i][j] = Slopes[i][j];
      Slopes[i][j] = (Decay * Weights[i][j]);
    }
  }
}

/*****************************************************************************/

/*
 *  BACKWARD_PASS():
 *  Goal is a vector of desired values for the output units.  Propagate
 *  the error back through the net, accumulating weight deltas.
 */

void BACKWARD_PASS(float *goal)
{
  int i, j, cix, *connect, unit_type;  /* cix is "connection index" */
  float errors, *slope, *weight;

  /* Compute error sums for output and other units  */

  for (i = FirstHidden; i < FirstOutput; i++)
    ErrorSums[i]=0.0;

  for (; i< Nunits; i++)
    ErrorSums[i] = ERRFUN(goal[i-FirstOutput], Outputs[i]);

  /* Back-prop.  When we reach a given unit in loop, error from all later
     units will have been collected. */

  for (j = Nunits-1; j >= FirstHidden; j--) {
    connect = Connections[j];
    weight = Weights[j];
    slope = Slopes[j];

    if(j < FirstOutput) 
      unit_type = HiddenUnitType;
    else
      unit_type = OutputUnitType;

    errors = ACTIVATION_PRIME(Outputs[j], unit_type) * ErrorSums[j];

    i = Nconnections[j];

    while (i > 0) {
      i--;
      cix = connect[i];
      ErrorSums[cix] += (errors * weight[i]);
      slope[i] += (errors * Outputs[cix]);
    }
  }
}

/*****************************************************************************/

/*
 *  ACTIVATION_PRIME():
 *  Given the unit's activation value and sum of weighted inputs,
 *  compute the derivative of the activation with respect to the sum.
 *  Defined types are SIGMOID (-1 to +1) and ASYMSIGMOID (0 to +1).
 */

float ACTIVATION_PRIME(float value, int unit_type)
{
  switch(unit_type)
    {
    case LINEAR:
      {
	return(1.);
      }
    case SIGMOID:
      {
	/* Symmetrical sigmoid function. */

	return (SigmoidPrimeOffset + (0.25 -  value*value));
      }
    default:
      {
	dbassert(unit_type==ASYMSIGMOID);
      
	/* asymmetrical sigmoid function in range 0.0 to 1.0. */
	
	return (SigmoidPrimeOffset + (value * (1.0 - value)));
      }
    }
}

/*****************************************************************************/

/*
 *  UPDATE_WEIGHTS():
 *  Update all weights in the network as a function of each weight's current
 *  slope, previous slope, and the size of the last jump.
 */

void UPDATE_WEIGHTS(void)
{
  int i, j;
  float next_step, shrink_factor, epsilon;

  shrink_factor = MaxFactor / (1.0 + MaxFactor);

  WeightCount = 0;
  DidGradient = 0;

  for (j = FirstHidden; j < Nunits; j++) {
    if (SplitEpsilon)
      epsilon = Epsilon / (float)Nconnections[j];
    else
      epsilon = Epsilon;

    for (i = 0; i < Nconnections[j]; i++) {
      next_step=0.0;
      WeightCount++;

      if ((DeltaWeights[j][i] > ModeSwitchThreshold) && UseQuickProp==1) {
	/* Last step was signif. +ive..... */

	if (Slopes[j][i] > 0.0)  /* Add in epsilon if +ive slope */
	  next_step +=(epsilon * Slopes[j][i]);

	/* If slope > (or close to) prev slope, take max size step. */

	if (Slopes[j][i] > (shrink_factor * PrevSlopes[j][i]))
	  next_step +=(MaxFactor * DeltaWeights[j][i]);
	else        /*  Use quadratic estimate */
	  next_step +=((Slopes[j][i]/(PrevSlopes[j][i]-Slopes[j][i]))
		       * DeltaWeights[j][i]);
      } else if ((DeltaWeights[j][i] < -ModeSwitchThreshold) && 
		 UseQuickProp==1) {

	/* Last step was signif. -ive.... */

	if (Slopes[j][i] < 0.0)/* Add in epsilon if -ive slope */
	  next_step +=(epsilon * Slopes[j][i]);

	/* If slope < (or close to) prev slope, take max size step. */

	if (Slopes[j][i] < (shrink_factor * PrevSlopes[j][i]))
	  next_step +=(MaxFactor * DeltaWeights[j][i]);
	else        /*  Use quadratic estimate */
	  next_step +=((Slopes[j][i]/(PrevSlopes[j][i]-Slopes[j][i]))
		       * DeltaWeights[j][i]);
      } else {

	/* Normal gradient descent, complete with momentum */

	DidGradient++;
	next_step +=((epsilon * Slopes[j][i])
		     + (Momentum * DeltaWeights[j][i]));
      }

      /* Set delta weight, and adjust the weight itself. */

      DeltaWeights[j][i] = next_step;
      Weights[j][i] += next_step;
    }
  }
}

/*****************************************************************************/

/*
 *  PRINT_RESULTS():
 *  Display line with error information.
 */


void PRINT_RESULTS(char *label, int nPatterns, int errorBits, float SSE)
{
#define printargs\
  " %s: %6.2f %%correct ;  RMSErr = %7.5f", label,\
    (float) 100*(1-(float)errorBits/(nPatterns*Noutputs)),\
    rms(SSE, nPatterns, Noutputs)
  printboth; 
#undef printargs

#define printargs "\n"
  printboth; 
#undef printargs
}

/*****************************************************************************/

/*
 *  PRINT_BEST():
 *  Display the best epoch of training.
 */

void PRINT_BEST(void)
{
#define printargs "*----------------------------------------------------*\n"
  printboth; 
#undef printargs

#define printargs "Best epoch: %d (TESTING: RMSErr = %.5f)\n",\
  BestEpoch, BestTestRMS
    printboth;
#undef printargs

#define printargs "*----------------------------------------------------*\n\n"
  printboth; 
#undef printargs
}

/*****************************************************************************/

/*
 *  TEST_SAVE():
 *  Save predictions to a .prd file
 */

void TEST_SAVE(void) 
{
  FILE  *outfile;
  int  i, j, m, n;
  char realfname[256];

  protected_file(realfname, fname, PRD_EXTENSION, ProtectFiles);
  outfile = fopen (realfname, "w"); 
  
  if(SaveBestWts && PredictOnly == 0) {
    for(m = 0; m < Nunits; m++)
      for(n = 0; n < Nunits; n++)
	Weights[m][n]=BestWeights[m][n];
  }
  
  for(j = 0; j < NTestingPatterns; j++) {
    FORWARD_PASS(TestInputs[j]);
    
    for (i = 0; i < Noutputs; i++)
      fprintf(outfile, "%f ", Outputs[FirstOutput+i]);
    
    for (i = FirstOutput; i < Nunits - 1; i++)
      fprintf(outfile, "%f ", TestOutputs[j][i-FirstOutput]);

    fprintf(outfile, "%f\n", TestOutputs[j][Nunits - 1 - FirstOutput]);
  }
  
  fclose (outfile);
  
#define printargs "Predictions saved to \"%s\"\n\n", realfname
  printboth;
#undef printargs
}

/*****************************************************************************/

/*
 *  DUMP_WEIGHTS():
 *  Save weights to a .wts file -- depending on settings, either
 *  the final trained weights, or best generalizing weights found. -phg
 */

void DUMP_WEIGHTS(void)
{
  FILE  *outfile;
  int  i, j, m, n;
  char realfname[256];
  
  protected_file(realfname, fname, WTS_EXTENSION, ProtectFiles);
  outfile = fopen (realfname, "w"); 
 
  EpochSaved = Epoch;

  if (SaveBestWts) {
    EpochSaved = BestEpoch;
    for (m = 0; m < Nunits; m++)
      for (n = 0; n < Nunits; n++)
	Weights[m][n]=BestWeights[m][n];
  }

  /* Indicate Epoch at which wts were frozen */
  fprintf(outfile, "EpochSaved %d\n", EpochSaved);

  for (j = FirstHidden; j < Nunits; j++)
    for (i = 0; i < Nconnections[j]; i++)
      /* Put each connection and weight on unique line  -phg */
      fprintf(outfile, "%d %d %.8f\n", j, Connections[j][i], Weights[j][i]);

  fclose (outfile);
   
#define printargs "Weights saved to \"%s\"\n", realfname
  printboth;
#undef printargs
}

/*****************************************************************************/

void FREE_MEMORY(void) 
{
  int i;

  for(i=0;i<NTrainingPatterns;i++) {
    free((void *)TrainingInputs[i]);
    free((void *)TrainingOutputs[i]);
  }

  free((void *)TrainingInputs);
  free((void *)TrainingOutputs);

  for(i=0;i<NTestingPatterns;i++) {
    free((void *)TestInputs[i]);
    free((void *)TestOutputs[i]);
  }

  free((void *)TestInputs);
  free((void *)TestOutputs);

  for(i=0;i<Nunits;i++) {
    free((void *)Connections[i]);
    free((void *)Weights[i]);
    free((void *)BestWeights[i]);
    free((void *)DeltaWeights[i]);
    free((void *)Slopes[i]);
    free((void *)PrevSlopes[i]);
  }

  free((void *)Connections);
  free((void *)Weights);
  free((void *)BestWeights);
  free((void *)Slopes);
  free((void *)PrevSlopes);
  free((void *)DeltaWeights);
  
  free((void *)Outputs);
  free((void *)ErrorSums);
  free((void *)Nconnections);
}

/*****************************************************************************/

void get_parameters(char *file_name, PARAMETERS *parameter_table, 
		    int n_parameters)
{
  FILE *fp;
  char line[256];

  fp = fopen(file_name, "r");

  while(fscanf(fp, "%[^\n]\n", &line[0]) != EOF) {
    process_line(line, parameter_table, n_parameters, 1);
  }
  
  fclose(fp);
}

/*****************************************************************************/

int process_line(char *line, 
		 PARAMETERS *parameter_table, int n_parameters,
		 int alert_flag)
{
  int i, k = -1;
  char *separators = " \t\v\f\r\n,", *first, *value;
  int ninputs, nhidden, noutputs;
  int c1, c2, c3, c4;

  if(line[0] == '#' || line[0] == '\n')
    return(1);

  first = strtok(line, separators);
  
  for(i = 0; i < n_parameters; i++)
    if(strcmp(first, parameter_table[i].keyword) == 0) {
      k = i; i = n_parameters;
    }

  if(k >= 0) {
    value = strtok(NULL, separators);

    switch(parameter_table[k].var_type)
      {
      case SHORT:
	{
	  *(short *) parameter_table[k].var_ptr = atoi(value);
	  break;
	}
      case INT:
	{
	  *(int *) parameter_table[k].var_ptr = atoi(value);
	  break;
	}
      case FLT:
	{
	  *(float *) parameter_table[k].var_ptr = (float) atof(value);
	  break;
	}
      case DBL:
	{
	  *(double *) parameter_table[k].var_ptr = atof(value);
	  break;
	}
      case STR:
	{
	  if(value != NULL)
	    strcpy((char *) parameter_table[k].var_ptr, value);
	  break;
	}
      }
  } else {

    /* Special Backprop NN code */
    /* Configuration token must appear BEFORE any ConnectCall tokens */

    if(strcmp(first, "ConnectCall") == 0) {
      c1 = atoi(strtok(NULL, separators));
      c2 = atoi(strtok(NULL, separators));
      c3 = atoi(strtok(NULL, separators));
      c4 = atoi(strtok(NULL, separators));
      
      CONNECT_LAYERS(c1, c2, c3, c4);

    } else if(strcmp(first, "Configuration") == 0) {
      ninputs = atoi(strtok(NULL, separators));
      nhidden = atoi(strtok(NULL, separators));
      noutputs = atoi(strtok(NULL, separators));

      BUILD_DATA_STRUCTURES(ninputs, nhidden, noutputs);

    } else {
      if(alert_flag == 1) 
	fprintf(stderr, "Did not recognize '%s' parameter\n", first);
    }
  }

  return(1);
}

/*****************************************************************************/

int count_lines(char *file_name)
{
  FILE *fp;
  int count = 0;
  char str[STR_MAX];

  fp = fopen(file_name, "r");

  if(fp == NULL)
    return(-1);

  while(fscanf(fp, "%[^\n]\n", &str[0]) != EOF)
    count++;

  fclose(fp);

  return(count);
}

/*****************************************************************************/

int file_exist(char *file_name)
{
  int flag;
  FILE *fp;

  fp = fopen(file_name, "r");

  if(fp != NULL) {
    flag = 1;
    fclose(fp);
  } else
    flag = 0;

  return(flag);
}

/*****************************************************************************/

void display_parameters(PARAMETERS *parameter_table, int n_parameters,
			FILE *fp)
{
  int i, n;
  char format[256];
  FILE *fp_temp;
  double dbl_tmp;
  float flt_tmp;

  if(fp == NULL)
    fp_temp = stdout;
  else
    fp_temp = fp;

  for(i = 0; i < n_parameters; i++) {
    switch(parameter_table[i].var_type)
      {
      case SHORT:
	{
	  fprintf(fp_temp, "%s %d\n",
		  parameter_table[i].keyword,
		  *(short *) parameter_table[i].var_ptr);
	  break;
	}
      case INT:
	{
	  fprintf(fp_temp, "%s %d\n",
		  parameter_table[i].keyword,
		  *(int *) parameter_table[i].var_ptr);
	  break;
	}
      case DBL:
	{
	  dbl_tmp = *(double *) parameter_table[i].var_ptr;

	  if(dbl_tmp > -0.0000000001 && dbl_tmp < 0.0000000001)
	    n = 2;
	  else if(log10(dbl_tmp) > 0.) 
	    n = 2;
	  else
	    n = (int) ceil(fabs(log10(dbl_tmp))) + 1;

	  sprintf(format, "%%s %%.%df\n", n);
	  fprintf(fp_temp, format,
		  parameter_table[i].keyword, dbl_tmp);
		  
	  break;
	}
      case FLT:
	{
	  flt_tmp = *(float *) parameter_table[i].var_ptr;

	  if(flt_tmp > -0.0000000001 && flt_tmp < 0.0000000001)
	    n = 2;
	  else if(log10(flt_tmp) > 0.) 
	    n = 2;
	  else
	    n = (int) ceil(fabs(log10((double) flt_tmp))) + 1;

	  sprintf(&format[0], "%%s %%.%df\n", n);
	  fprintf(fp_temp, format, 
		  parameter_table[i].keyword, flt_tmp);
	  break;
	}
      case STR:
	{
	  fprintf(fp_temp, "%s %s\n",
		  parameter_table[i].keyword,
		  (char *) parameter_table[i].var_ptr);
	  break;
	}
      }
  }
}

/*****************************************************************************/

void  protected_file(char *realfname, char *fname, 
		     char *extension, int protect_flag) 
{
  int flag = -1, count = 1;

  sprintf(realfname, "%s.%s", fname, extension);
  
  if (protect_flag > 0 && file_exist(realfname)) {

    while(flag < 0) {
      sprintf(realfname, "%s-%d.%s", fname, count, extension);
      if(file_exist(realfname)) {
	count++;
      } else {
	flag = 1;
      } 
    }
  }
  
  return;
}

/*****************************************************************************/

float rms(float sse, int n, int nout) 
{
  return((float) sqrt((double) (sse/((double) (n*nout)))));
}

/*****************************************************************************/

void checkEOF(int scanfReturn)
{
  if (scanfReturn == EOF || scanfReturn == 0) { 
    fprintf(stderr,
	    "ERROR: Premature end-of-file encountered on input.\n");
    exit(1);
  }
}

/*****************************************************************************/
